from typing import List, Union
from azure.durable_functions.models.ReplaySchema import ReplaySchema
from .orchestrator_test_utils \
    import get_orchestration_state_result, assert_orchestration_state_equals, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.RetryOptions import RetryOptions
from azure.durable_functions.models.actions.CallActivityWithRetryAction \
    import CallActivityWithRetryAction


RETRY_OPTIONS = RetryOptions(5000, 3)


def generator_function(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = yield context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")
    task3 = yield context.call_activity_with_retry(
        "Hello",  retry_options, "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs

def generator_function_try_catch(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    result = None
    try:
        result = yield context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    except:
        result = yield context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")
    return result

def generator_function_concurrent_retries(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")
    task3 = context.call_activity_with_retry(
        "Hello",  retry_options, "London")

    outputs = yield context.task_all([task1, task2, task3])

    return outputs

def generator_function_two_concurrent_retries_when_all(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")

    outputs = yield context.task_all([task1, task2])

    return outputs

def generator_function_two_concurrent_retries_when_any(context):
    outputs = []

    retry_options = RETRY_OPTIONS
    task1 = context.call_activity_with_retry(
        "Hello", retry_options, "Tokyo")
    task2 = context.call_activity_with_retry(
        "Hello",  retry_options, "Seattle")

    outputs = yield context.task_any([task1, task2])

    return outputs.result


def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)


def add_hello_action(state: OrchestratorState, input_: Union[List[str], str]):
    retry_options = RETRY_OPTIONS
    actions = []
    inputs = input_
    if not isinstance(input_, list):
        inputs = [input_]
    for input_ in inputs:
        action = CallActivityWithRetryAction(
            function_name='Hello', retry_options=retry_options, input_=input_)
        actions.append(action)
    state._actions.append(actions)


def add_hello_failed_events(
        context_builder: ContextBuilder, id_: int, reason: str, details: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_failed_event(
        id_=id_, reason=reason, details=details)


def add_hello_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=result)


def add_retry_timer_events(context_builder: ContextBuilder, id_: int):
    fire_at = context_builder.add_timer_created_event(id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_timer_fired_event(id_=id_, fire_at=fire_at)

def add_two_retriable_events_completing_out_of_order(context_builder: ContextBuilder,
        failed_reason, failed_details):
    ## Schedule tasks
    context_builder.add_task_scheduled_event(name='Hello', id_=0) # Tokyo task
    context_builder.add_task_scheduled_event(name='Hello', id_=1) # Seattle task

    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()

    ## Task failures and timer-scheduling

    # tasks fail "out of order"
    context_builder.add_task_failed_event(
        id_=1, reason=failed_reason, details=failed_details) # Seattle task
    fire_at_1 = context_builder.add_timer_created_event(2) # Seattle timer

    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()

    context_builder.add_task_failed_event(
        id_=0, reason=failed_reason, details=failed_details) # Tokyo task
    fire_at_2 = context_builder.add_timer_created_event(3) # Tokyo timer

    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()

    ## fire timers
    context_builder.add_timer_fired_event(id_=2, fire_at=fire_at_1) # Seattle timer
    context_builder.add_timer_fired_event(id_=3, fire_at=fire_at_2) # Tokyo timer

    ## Complete events
    context_builder.add_task_scheduled_event(name='Hello', id_=4) # Seattle task
    context_builder.add_task_scheduled_event(name='Hello', id_=5) # Tokyo task

    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=4, result="\"Hello Seattle!\"")
    context_builder.add_task_completed_event(id_=5, result="\"Hello Tokyo!\"")


def test_initial_orchestration_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_retry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_timer_entry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_failed_retry():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_with_failed_retry_timer_added():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_successful_tokyo_with_failed_retry_timer_added():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_completed_events(context_builder, 2, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_tokyo_hit_max_attempts():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_hello_failed_events(context_builder, 4, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 5)

    try:
        result = get_orchestration_state_result(
            context_builder, generator_function)
        # expected an exception
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)

        expected_state = base_expected_state()
        add_hello_action(expected_state, 'Tokyo')

        error_msg = f'{failed_reason} \n {failed_details}'
        expected_state._error = error_msg
        state_str = expected_state.to_json_string()
        
        expected_error_str = f"{error_msg}{error_label}{state_str}"
        assert expected_error_str == error_str

def test_failed_tokyo_hit_max_attempts_in_try_catch():
    # This test ensures that APIs can still be invoked after a failed CallActivityWithRetry invocation
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_simple_function')
    
    # events for first task: "Hello Tokyo"
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 1)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_hello_failed_events(context_builder, 4, failed_reason, failed_details)
    # we have an "extra" timer to wait for, due to legacy behavior in DTFx.
    add_retry_timer_events(context_builder, 5)

    # events to task in except block
    add_hello_completed_events(context_builder, 6, "\"Hello Seattle!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_try_catch)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_hello_action(expected_state, 'Seattle')
    expected_state._output = "Hello Seattle!"
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_concurrent_retriable_results():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_concurrent_retriable')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_hello_failed_events(context_builder, 1, failed_reason, failed_details)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_retry_timer_events(context_builder, 4)
    add_retry_timer_events(context_builder, 5)
    add_hello_completed_events(context_builder, 6, "\"Hello Tokyo!\"")
    add_hello_completed_events(context_builder, 7, "\"Hello Seattle!\"")
    add_hello_completed_events(context_builder, 8, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_concurrent_retries)

    expected_state = base_expected_state()
    add_hello_action(expected_state, ['Tokyo', 'Seattle', 'London'])
    expected_state._output = ["Hello Tokyo!", "Hello Seattle!", "Hello London!"]
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_concurrent_retriable_results_unordered_arrival():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_concurrent_retriable_unordered_results')
    add_hello_failed_events(context_builder, 0, failed_reason, failed_details)
    add_hello_failed_events(context_builder, 1, failed_reason, failed_details)
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_retry_timer_events(context_builder, 4)
    add_retry_timer_events(context_builder, 5)
    # events arrive in non-sequential different order
    add_hello_completed_events(context_builder, 8, "\"Hello London!\"")
    add_hello_completed_events(context_builder, 6, "\"Hello Tokyo!\"")
    add_hello_completed_events(context_builder, 7, "\"Hello Seattle!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_concurrent_retries)

    expected_state = base_expected_state()
    add_hello_action(expected_state, ['Tokyo', 'Seattle', 'London'])
    expected_state._output = ["Hello Tokyo!", "Hello Seattle!", "Hello London!"]
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_concurrent_retriable_results_mixed_arrival():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_concurrent_retriable_unordered_results')
    # one task succeeds, the other two fail at first, and succeed on retry
    add_hello_failed_events(context_builder, 1, failed_reason, failed_details)
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_failed_events(context_builder, 2, failed_reason, failed_details)
    add_retry_timer_events(context_builder, 3)
    add_retry_timer_events(context_builder, 4)
    add_hello_completed_events(context_builder, 6, "\"Hello London!\"")
    add_hello_completed_events(context_builder, 5, "\"Hello Seattle!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_concurrent_retries)

    expected_state = base_expected_state()
    add_hello_action(expected_state, ['Tokyo', 'Seattle', 'London'])
    expected_state._output = ["Hello Tokyo!", "Hello Seattle!", "Hello London!"]
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_concurrent_retriable_results_alternating_taskIDs_when_all():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_concurrent_retriable_unordered_results')

    add_two_retriable_events_completing_out_of_order(context_builder, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function_two_concurrent_retries_when_all)

    expected_state = base_expected_state()
    add_hello_action(expected_state, ['Tokyo', 'Seattle'])
    expected_state._output = ["Hello Tokyo!", "Hello Seattle!"]
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_concurrent_retriable_results_alternating_taskIDs_when_any():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    context_builder = ContextBuilder('test_concurrent_retriable_unordered_results')

    add_two_retriable_events_completing_out_of_order(context_builder, failed_reason, failed_details)

    result = get_orchestration_state_result(
        context_builder, generator_function_two_concurrent_retries_when_any)

    expected_state = base_expected_state()
    add_hello_action(expected_state, ['Tokyo', 'Seattle'])
    expected_state._output = "Hello Seattle!"
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


"""Setup for the durable function module."""
import pathlib
import os
import shutil
import subprocess
import sys

from glob import glob
from setuptools import setup, find_packages
from distutils.command import build

with open("README.md", "r",  encoding="utf8") as fh:
    long_description = fh.read()

class BuildModule(build.build):
    """Used to build the module."""

    def run(self, *args, **kwargs):
        """Execute the build.

        :param args:
        :param kwargs:
        """
        super().run(*args, **kwargs)


setup(
    name='azure-functions-durable',
    packages=find_packages(exclude=[
        "tests",
        "samples",
        "scripts",
        "azure"
    ]),
    use_scm_version=True,
    setup_requires=['setuptools_scm'],
    author="Azure Functions team at Microsoft Corp.",
    author_email="azurefunctions@microsoft.com",
    keywords="azure functions azurefunctions python serverless workflows durablefunctions",
    url="https://github.com/Azure/azure-functions-durable-python",
    description='Durable Functions For Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    classifiers=[
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Programming Language :: Python :: 3',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Operating System :: MacOS :: MacOS X',
        'Environment :: Web Environment',
        'Development Status :: 5 - Production/Stable',
    ],
    license='MIT',
    python_requires='>=3.6,<4',
    install_requires=[
        'azure-functions>=1.12.0',
        'aiohttp>=3.6.2',
        'requests==2.*',
        'python-dateutil>=2.8.0',
        'furl>=2.1.0'
    ],
    extra_requires=[
        'flake8==3.7.8',
        'flake8-docstrings==1.5.0',
        'pytest==7.1.2',
        'python-dateutil==2.8.0',
        'requests==2.22.0',
        'jsonschema==3.2.0',
        'aiohttp==3.6.2',
        'azure-functions>=1.2.0',
        'nox==2019.11.9',
        'furl==2.1.0',
        'pytest-asyncio==0.20.2'
    ],
    include_package_data=True,
    data_files= [
        ('_manifest', list(filter(os.path.isfile, glob('_manifest/**/*', recursive=True)))),
    ],
    cmdclass={
        'build': BuildModule
    },
    test_suite='tests'
)



import json
import random

random.seed(10)

def main(phoneNumber: str, message):
  payload = {
    "body": f"Hey! You may want to check on your repo, there are too many open issues",
    "to": phoneNumber
  }

  message.set(json.dumps(payload))
  return "Message sent!"



import logging

import azure.functions as func
import azure.durable_functions as df
from ..shared_code.MyClasses import SerializableClass

def orchestrator_function(context: df.DurableOrchestrationContext):
    """This function provides the core function chaining orchestration logic

    Parameters
    ----------
    context: DurableOrchestrationContext
        This context has the past history and the durable orchestration API's to
        create orchestrations

    Returns
    -------
    int
        The number contained in the SerializableClass input object
    """
    input_: SerializableClass = context.get_input()
    num1: int = input_.show_number()

    # The custom class is also correctly serialized when calling an activity
    num2 = yield context.call_activity("DurableActivity", SerializableClass(5))
    return num1

main = df.Orchestrator.create(orchestrator_function)



from typing import List, Optional, Dict, Any
from .Signal import Signal
from azure.functions._durable_functions import _serialize_custom_object
from .OperationResult import OperationResult
import json


class EntityState:
    """Entity State.

    Used to communicate the state of the entity back to the durable extension
    """

    def __init__(self,
                 results: List[OperationResult],
                 signals: List[Signal],
                 entity_exists: bool = False,
                 state: Optional[str] = None):
        self.entity_exists = entity_exists
        self.state = state
        self._results = results
        self._signals = signals

    @property
    def results(self) -> List[OperationResult]:
        """Get list of results of the entity.

        Returns
        -------
        List[OperationResult]:
            The results of the entity
        """
        return self._results

    @property
    def signals(self) -> List[Signal]:
        """Get list of signals to the entity.

        Returns
        -------
        List[Signal]:
            The signals of the entity
        """
        return self._signals

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Any] = {}
        # Serialize the OperationResult list
        serialized_results = list(map(lambda x: x.to_json(), self.results))

        json_dict["entityExists"] = self.entity_exists
        json_dict["entityState"] = json.dumps(self.state, default=_serialize_custom_object)
        json_dict["results"] = serialized_results
        json_dict["signals"] = self.signals
        return json_dict

    def to_json_string(self) -> str:
        """Convert object into a json string.

        Returns
        -------
        str
            The instance of the object in json string format
        """
        # TODO: Same implementation as in Orchestrator.py, we should refactor to shared a base
        json_dict = self.to_json()
        return json.dumps(json_dict)



import nox

@nox.session(python=["3.7","3.8"])
def tests(session):
    # same as pip install -r -requirements.txt
    session.install("-r", "requirements.txt")
    session.install("pytest")
    session.run("pytest", "-v", "tests")


@nox.session(python=["3.7", "3.8"])
def lint(session):
    session.install("flake8")
    session.install("flake8-docstrings")
    session.run("flake8", "./azure/")

@nox.session(python=["3.7", "3.8"])
def typecheck(session):
    session.install("-r", "requirements.txt")
    session.install("mypy")
    session.run("mypy", "./azure/")

@nox.session(python=["3.7", "3.8"])
def autopep(session):
    session.install("-r", "requirements.txt")
    session.run("autopep8", "--in-place --aggressive --aggressive --recursive \"./azure/\"")


from azure.durable_functions.decorators.metadata import OrchestrationTrigger,\
    EntityTrigger, ActivityTrigger, DurableClient
from azure.durable_functions.constants import ORCHESTRATION_TRIGGER, \
    ACTIVITY_TRIGGER, ENTITY_TRIGGER, DURABLE_CLIENT
from azure.functions.decorators.core import BindingDirection


def test_OrchestrationTrigger_with_full_params():
    trigger = OrchestrationTrigger(
        name="myContext",
        orchestration="myOrchestratorFunction"
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ORCHESTRATION_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "orchestrationTrigger",
        "name": "myContext",
        "orchestration": "myOrchestratorFunction"
    }

def test_OrchestrationTrigger_with_default_params():
    trigger = OrchestrationTrigger(
        name="myContext",
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ORCHESTRATION_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "orchestrationTrigger",
        "name": "myContext"
    }

def test_EntityTrigger_with_full_params():
    trigger = EntityTrigger(
        name="myContext",
        entity_name="myEntityFunction"
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ENTITY_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "entityTrigger",
        "name": "myContext",
        "entityName": "myEntityFunction"
    }

def test_EntityTrigger_with_default_params():
    trigger = EntityTrigger(
        name="myContext",
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ENTITY_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "entityTrigger",
        "name": "myContext"
    }

def test_ActivityTrigger_with_full_params():
    trigger = ActivityTrigger(
        name="myInput",
        activity="myActivityFunction"
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ACTIVITY_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "activityTrigger",
        "name": "myInput",
        "activity": "myActivityFunction"
    }

def test_ActivityTrigger_with_default_params():
    trigger = ActivityTrigger(
        name="myInput"
    )

    binding_name = trigger.get_binding_name()
    assert binding_name == ACTIVITY_TRIGGER

    dict_repr = trigger.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "activityTrigger",
        "name": "myInput"
    }

def test_DurableClient_with_full_params():
    input_binding = DurableClient(
        name="myClient",
        task_hub="myTaskHub",
        connection_name="myConnection"
    )

    binding_name = input_binding.get_binding_name()
    assert binding_name == DURABLE_CLIENT

    dict_repr = input_binding.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "durableClient",
        "name": "myClient",
        "taskHub": "myTaskHub",
        "connectionName": "myConnection"
    }

def test_DurableClient_with_default_params():
    input_binding = DurableClient(
        name="myClient",
    )

    binding_name = input_binding.get_binding_name()
    assert binding_name == DURABLE_CLIENT

    dict_repr = input_binding.get_dict_repr()
    assert dict_repr == {
        "direction": BindingDirection.IN,
        "type": "durableClient",
        "name": "myClient"
    }


import logging
import json

import azure.functions as func
import azure.durable_functions as df


def entity_function(context: df.DurableEntityContext):
    """A Counter Durable Entity.

    A simple example of a Durable Entity that implements
    a simple counter.

    Parameters
    ----------
    context (df.DurableEntityContext):
        The Durable Entity context, which exports an API
        for implementing durable entities.
    """

    current_value = context.get_state(lambda: 0)
    operation = context.operation_name
    if operation == "add":
        amount = context.get_input()
        current_value += amount
    elif operation == "reset":
        current_value = 0
    elif operation == "get":
        pass
    
    context.set_state(current_value)
    context.set_result(current_value)


main = df.Entity.create(entity_function)


import os

# The imports below are only needed when the logic in get_run_logs
# is reapplied.
# from azureml.core import Experiment, Workspace
# from azureml.pipeline.core import PipelineRun

# from ..shared.auth_helper import get_service_principal_auth

_SUBSCRIPTION_ID_ENV_NAME = "SubscriptionId"
_RESOURCE_GROUP_NAME_ENV_NAME = "ResourceGroupName"
_AML_WORKSPACE_NAME_ENV_NAME = "AMLWorkspaceName"

_RUN_URL_TEMPLATE = (
    "https://mlworkspace.azure.ai/portal/subscriptions/{0}"
    "/resourceGroups/{1}/providers/Microsoft.MachineLearningServices"
    "/workspaces/{2}/experiments/{3}/runs/{4}"
)

def get_run_url_from_env(run_id: str, experiment_name: str):
    """Retrieves the appropriate environment variables.
       Uses an run url template and formats that string with the appropriate parameters.
       Return a string containing an run url based on
       function params and environment variables.

    Arguments:
        run_id string -- The string representation of the experiments run id
        experiment_name string -- The string representation of the experiments name

    Returns:
        string -- The url string that points to the current run params.
    """
    if not run_id or not experiment_name:
        raise ValueError("Missing required param")
    subscription_id = os.environ.get(_SUBSCRIPTION_ID_ENV_NAME)
    resource_group_name = os.environ.get(_RESOURCE_GROUP_NAME_ENV_NAME)
    aml_workspace_name = os.environ.get(_AML_WORKSPACE_NAME_ENV_NAME)

    return _RUN_URL_TEMPLATE.format(subscription_id, resource_group_name, \
        aml_workspace_name, experiment_name, run_id)


def get_run_logs(run_id: str, experiment_name: str):
    """Retrieves the appropriate environment variables.
       Retrieves steps for the experiments pipeline run.
       Builds a dictionary of logs for each step by the steps id.

    Arguments:
        run_id string -- The string representation of the experiments run id
        experiment_name string -- The string representation of the experiments name

    Returns:
        dictionary -- A dictionary containing logs for pipeline run
        steps keyed by the steps run id.
    """

    if not run_id or not experiment_name:
        raise ValueError("Missing required param")

    # Commenting out code due to a bug in the adf pipeline that doesn't allow
    # us to properly set the aml sdk version. This bug is results in the
    # PipelineRun to return Run objects instead of StepRun objects when
    # the get_steps object is called. Since Run objects don't have the
    # get_job_log method, the code errors out when running via adf, but not
    # when testing locally. When this bug is resolved, this code can be un
    # commented and redeployed.

    # svc_pr = get_service_principal_auth()
    # workspace = Workspace(
    #     subscription_id=os.environ[_SUBSCRIPTION_ID_ENV_NAME],
    #     resource_group=os.environ[_RESOURCE_GROUP_NAME_ENV_NAME],
    #     workspace_name=os.environ[_AML_WORKSPACE_NAME_ENV_NAME],
    #     auth=svc_pr)

    # experiment = Experiment(workspace, experiment_name)
    # pipeline_run = PipelineRun(experiment, run_id)
    # run_steps = list(pipeline_run.get_steps())

    # iterate over steps to get logs
    run_log_dict = dict()
    # for step in run_steps:
    #     j_log = step.get_job_log()
    #     run_log_dict[str(step.id)] = j_log

    return run_log_dict



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from tests.test_utils.ContextBuilder import ContextBuilder
from .orchestrator_test_utils \
    import get_orchestration_property, assert_orchestration_state_equals, assert_valid_schema
from azure.durable_functions.models.actions.CreateTimerAction import CreateTimerAction
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.constants import DATETIME_STRING_FORMAT
from datetime import datetime, timedelta, timezone

def generator_function(context):
    # Create a timezone aware datetime object, just like a normal
    # call to `context.current_utc_datetime` would create
    timestamp = "2020-07-23T21:56:54.936700Z"
    deadline = datetime.strptime(timestamp, DATETIME_STRING_FORMAT)
    deadline = deadline.replace(tzinfo=timezone.utc)

    for _ in range(0, 3):
        deadline = deadline + timedelta(seconds=30)
        yield context.create_timer(deadline)

def generator_function_compound_task(context):
    # Create a timezone aware datetime object, just like a normal
    # call to `context.current_utc_datetime` would create
    timestamp = "2020-07-23T21:56:54.936700Z"
    deadline = datetime.strptime(timestamp, DATETIME_STRING_FORMAT)
    deadline = deadline.replace(tzinfo=timezone.utc)

    tasks = []
    for _ in range(0, 3):
        deadline = deadline + timedelta(seconds=30)
        tasks.append(context.create_timer(deadline))
    yield context.task_any(tasks)

def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)

def add_timer_fired_events(context_builder: ContextBuilder, id_: int, timestamp: str,
        is_played: bool = True):
    fire_at: str = context_builder.add_timer_created_event(id_, timestamp)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_timer_fired_event(id_=id_, fire_at=fire_at, is_played=is_played)

def add_timer_action(state: OrchestratorState, fire_at: datetime):
    action = CreateTimerAction(fire_at=fire_at)
    state._actions.append([action])

def test_is_replaying_initial_value():

    context_builder = ContextBuilder("", is_replaying=False)
    result = get_orchestration_property(
        context_builder, generator_function, "durable_context")

    assert result.is_replaying == False

def test_is_replaying_one_replayed_event():

    timestamp = "2020-07-23T21:56:54.9367Z"
    fire_at = datetime.strptime(timestamp, DATETIME_STRING_FORMAT) + timedelta(seconds=30)
    fire_at_str = fire_at.strftime(DATETIME_STRING_FORMAT)

    context_builder = ContextBuilder("")
    add_timer_fired_events(context_builder, 0, fire_at_str, is_played=True)

    result = get_orchestration_property(
        context_builder, generator_function, "durable_context")

    assert result.is_replaying == True

def test_is_replaying_one_replayed_one_not():

    timestamp = "2020-07-23T21:56:54.9367Z"
    fire_at = datetime.strptime(timestamp, DATETIME_STRING_FORMAT) + timedelta(seconds=30)
    fire_at_str = fire_at.strftime(DATETIME_STRING_FORMAT)
    fire_at2 = datetime.strptime(timestamp, DATETIME_STRING_FORMAT) + timedelta(seconds=60)
    fire_at_str2 = fire_at2.strftime(DATETIME_STRING_FORMAT)

    context_builder = ContextBuilder("")
    add_timer_fired_events(context_builder, 0, fire_at_str, is_played=True)
    add_timer_fired_events(context_builder, 1, fire_at_str2, is_played=False)


    result = get_orchestration_property(
        context_builder, generator_function, "durable_context")

    assert result.is_replaying == False

def test_is_replaying_propagates_in_compound_task():

    timestamp = "2020-07-23T21:56:54.9367Z"
    fire_at = datetime.strptime(timestamp, DATETIME_STRING_FORMAT) + timedelta(seconds=30)
    fire_at_str = fire_at.strftime(DATETIME_STRING_FORMAT)

    context_builder = ContextBuilder("")
    add_timer_fired_events(context_builder, 0, fire_at_str, is_played=True)

    result = get_orchestration_property(
        context_builder, generator_function_compound_task, "durable_context")

    assert result.is_replaying == True




"""Model definitions for Durable Functions."""
from .DurableOrchestrationBindings import DurableOrchestrationBindings
from .DurableOrchestrationClient import DurableOrchestrationClient
from .DurableOrchestrationContext import DurableOrchestrationContext
from .OrchestratorState import OrchestratorState
from .OrchestrationRuntimeStatus import OrchestrationRuntimeStatus
from .PurgeHistoryResult import PurgeHistoryResult
from .RetryOptions import RetryOptions
from .DurableHttpRequest import DurableHttpRequest
from .TokenSource import ManagedIdentityTokenSource
from .DurableEntityContext import DurableEntityContext

__all__ = [
    'DurableOrchestrationBindings',
    'DurableOrchestrationClient',
    'DurableEntityContext',
    'DurableOrchestrationContext',
    'DurableHttpRequest',
    'ManagedIdentityTokenSource',
    'OrchestratorState',
    'OrchestrationRuntimeStatus',
    'PurgeHistoryResult',
    'RetryOptions'
]



import json
from typing import Dict, Optional

from azure.durable_functions.models.FunctionContext import FunctionContext


class DurableOrchestrationBindings:
    """Binding information.

    Provides information relevant to the creation and management of
    durable functions.
    """

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self, taskHubName: str, creationUrls: Dict[str, str],
                 managementUrls: Dict[str, str], rpcBaseUrl: Optional[str] = None, **kwargs):
        self._task_hub_name: str = taskHubName
        self._creation_urls: Dict[str, str] = creationUrls
        self._management_urls: Dict[str, str] = managementUrls
        # TODO: we can remove the Optional once we drop support for 1.x,
        # this is always provided in 2.x
        self._rpc_base_url: Optional[str] = rpcBaseUrl
        self._client_data = FunctionContext(**kwargs)

    @property
    def task_hub_name(self) -> str:
        """Get the name of the container that is used for orchestrations."""
        return self._task_hub_name

    @property
    def creation_urls(self) -> Dict[str, str]:
        """Get the URLs that are used for creating new orchestrations."""
        return self._creation_urls

    @property
    def management_urls(self) -> Dict[str, str]:
        """Get the URLs that are used for managing orchestrations."""
        return self._management_urls

    @property
    def rpc_base_url(self) -> Optional[str]:
        """Get the base url communication between out of proc workers and the function host."""
        return self._rpc_base_url

    @property
    def client_data(self) -> FunctionContext:
        """Get any additional client data provided within the context of the client."""
        return self._client_data

    @classmethod
    def from_json(cls, json_string):
        """Convert the value passed into a new instance of the class.

        Parameters
        ----------
        json_string
            Context passed a JSON serializable value to be converted into an
            instance of the class

        Returns
        -------
        DurableOrchestrationBindings
            New instance of the durable orchestration binding class
        """
        json_dict = json.loads(json_string)
        return cls(**json_dict)



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from .orchestrator_test_utils \
    import assert_orchestration_state_equals, get_orchestration_state_result, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.actions.CallActivityAction \
    import CallActivityAction
from azure.durable_functions.models.actions.ContinueAsNewAction \
    import ContinueAsNewAction


def generator_function(context):
    yield context.call_activity("Hello", "Tokyo")
    context.continue_as_new("Cause I can")


def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)


def add_hello_action(state: OrchestratorState, input_: str):
    action = CallActivityAction(function_name='Hello', input_=input_)
    state.actions.append([action])


def add_continue_as_new_action(state: OrchestratorState, input_: str):
    action = ContinueAsNewAction(input_=input_)
    state.actions.append([action])
    state._is_done = True


def add_hello_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_task_scheduled_event(name='Hello', id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=result)


def test_initial_orchestration_state():
    context_builder = ContextBuilder('test_simple_function')

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_tokyo_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_completed_events(context_builder, 0, "\"Hello Tokyo!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_hello_action(expected_state, 'Tokyo')
    add_continue_as_new_action(expected_state, 'Cause I can')
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)



from typing import Dict, Any


class PurgeHistoryResult:
    """Information provided when the request to purge history has been made."""

    # parameter names are as defined by JSON schema and do not conform to PEP8 naming conventions
    def __init__(self, instancesDeleted: int, **kwargs):
        self._instances_deleted: int = instancesDeleted
        if kwargs is not None:
            for key, value in kwargs.items():
                self.__setattr__(key, value)

    @classmethod
    def from_json(cls, json_obj: Dict[Any, Any]):
        """Convert the value passed into a new instance of the class.

        Parameters
        ----------
        json_obj: Dict[Any, Any]
            JSON object to be converted into an instance of the class

        Returns
        -------
        PurgeHistoryResult
            New instance of the durable orchestration status class
        """
        return cls(**json_obj)

    @property
    def instances_deleted(self):
        """Get the number of deleted instances."""
        return self._instances_deleted



from azure.durable_functions.models.ReplaySchema import ReplaySchema
import json

from azure.durable_functions.models import OrchestratorState
from azure.durable_functions.models.actions import CallActivityAction
from .orchestrator_test_utils import get_orchestration_state_result, \
    assert_orchestration_state_equals, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder


def generator_function(context):
    activity_count = yield context.call_activity("GetActivityCount")
    tasks = []
    for i in range(activity_count):
        current_task = context.call_activity("ParrotValue", str(i))
        tasks.append(current_task)
    values = yield context.task_all(tasks)
    results = yield context.call_activity("ShowMeTheSum", values)
    return results


def base_expected_state(output=None, error=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema)


def add_completed_event(
        context_builder: ContextBuilder, id_: int, name: str, result):
    context_builder.add_task_scheduled_event(name=name, id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_completed_event(id_=id_, result=json.dumps(result))


def add_failed_event(
        context_builder: ContextBuilder, id_: int, name: str, reason: str, details: str):
    context_builder.add_task_scheduled_event(name=name, id_=id_)
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_task_failed_event(
        id_=id_, reason=reason, details=details)


def add_completed_task_set_events(
        context_builder: ContextBuilder, start_id: int, name: str, volume: int,
        failed_index: int = -1, failed_reason: str = '', failed_details: str = ''):
    for i in range(volume):
        if i != failed_index:
            add_completed_event(context_builder, start_id + i, name, i)
        else:
            add_failed_event(context_builder, start_id + i, name, failed_reason, failed_details)


def add_single_action(state: OrchestratorState, function_name: str, input_):
    action = CallActivityAction(function_name=function_name, input_=input_)
    state.actions.append([action])


def add_multi_actions(state: OrchestratorState, function_name: str, volume: int):
    actions = []
    for i in range(volume):
        action = CallActivityAction(function_name=function_name, input_=json.dumps(i))
        actions.append(action)
    state.actions.append(actions)


def test_initial_call():
    context_builder = ContextBuilder('test_fan_out_fan_in_function')

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_single_action(expected_state, function_name='GetActivityCount', input_=None)
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_get_activity_count_success():
    activity_count = 5
    context_builder = ContextBuilder('test_fan_out_fan_in_function')
    add_completed_event(context_builder, 0, 'GetActivityCount', activity_count)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_single_action(expected_state, function_name='GetActivityCount', input_=None)
    add_multi_actions(expected_state, function_name='ParrotValue', volume=activity_count)
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_parrot_value_success():
    activity_count = 5
    context_builder = ContextBuilder('test_fan_out_fan_in_function')
    add_completed_event(context_builder, 0, 'GetActivityCount', activity_count)
    add_completed_task_set_events(context_builder, 1, 'ParrotValue', activity_count)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state()
    add_single_action(expected_state, function_name='GetActivityCount', input_=None)
    add_multi_actions(expected_state, function_name='ParrotValue', volume=activity_count)
    results = []
    for i in range(activity_count):
        results.append(i)
    add_single_action(expected_state, function_name='ShowMeTheSum', input_=results)
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_show_me_the_sum_success():
    activity_count = 5
    sum_ = 0
    for i in range(activity_count):
        sum_ += i
    sum_results = f"Well that's nice {sum_}!"
    context_builder = ContextBuilder('test_fan_out_fan_in_function')
    add_completed_event(context_builder, 0, 'GetActivityCount', activity_count)
    add_completed_task_set_events(context_builder, 1, 'ParrotValue', activity_count)
    add_completed_event(
        context_builder, activity_count + 1, 'ShowMeTheSum', sum_results)

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(sum_results)
    add_single_action(expected_state, function_name='GetActivityCount', input_=None)
    add_multi_actions(expected_state, function_name='ParrotValue', volume=activity_count)
    results = []
    for i in range(activity_count):
        results.append(i)
    add_single_action(expected_state, function_name='ShowMeTheSum', input_=results)
    expected_state._is_done = True
    expected = expected_state.to_json()

    assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)


def test_failed_parrot_value():
    failed_reason = 'Reasons'
    failed_details = 'Stuff and Things'
    activity_count = 5
    context_builder = ContextBuilder('test_fan_out_fan_in_function')
    add_completed_event(context_builder, 0, 'GetActivityCount', activity_count)
    add_completed_task_set_events(context_builder, 1, 'ParrotValue', activity_count,
                                  2, failed_reason, failed_details)

    try:
        result = get_orchestration_state_result(
            context_builder, generator_function)
        # we expected an exception
        assert False
    except Exception as e:
        error_label = "\n\n$OutOfProcData$:"
        error_str = str(e)

        expected_state = base_expected_state(error=f'{failed_reason} \n {failed_details}')
        add_single_action(expected_state, function_name='GetActivityCount', input_=None)
        add_multi_actions(expected_state, function_name='ParrotValue', volume=activity_count)

        error_msg = f'{failed_reason} \n {failed_details}'
        expected_state._error = error_msg
        state_str = expected_state.to_json_string()
        
        expected_error_str = f"{error_msg}{error_label}{state_str}"
        assert expected_error_str == error_str



from enum import IntEnum


class ActionType(IntEnum):
    """Defines the values associated to the types of activities that can be scheduled."""

    CALL_ACTIVITY: int = 0
    CALL_ACTIVITY_WITH_RETRY: int = 1
    CALL_SUB_ORCHESTRATOR: int = 2
    CALL_SUB_ORCHESTRATOR_WITH_RETRY: int = 3
    CONTINUE_AS_NEW: int = 4
    CREATE_TIMER: int = 5
    WAIT_FOR_EXTERNAL_EVENT: int = 6
    CALL_ENTITY = 7
    CALL_HTTP: int = 8
    SIGNAL_ENTITY: int = 9
    WHEN_ANY = 11
    WHEN_ALL = 12



from typing import Any, Dict

from .Action import Action
from .ActionType import ActionType
from ..DurableHttpRequest import DurableHttpRequest
from ..utils.json_utils import add_attrib, add_json_attrib


class CallHttpAction(Action):
    """Defines the structure of the Call Http object.

    Provides the information needed by the durable extension to be able to schedule the activity.
    """

    def __init__(self, http_request: DurableHttpRequest):
        self._action_type: int = ActionType.CALL_HTTP
        self.http_request = http_request

    @property
    def action_type(self) -> int:
        """Get the type of action this class represents."""
        return ActionType.CALL_HTTP

    def to_json(self) -> Dict[str, Any]:
        """Convert object into a json dictionary.

        Returns
        -------
        Dict[str, Any]
            The instance of the class converted into a json dictionary
        """
        json_dict: Dict[str, Any] = {}
        add_attrib(json_dict, self, 'action_type', 'actionType')
        add_json_attrib(json_dict, self, 'http_request', 'httpRequest')
        return json_dict



# This function is not intended to be invoked directly. Instead it will be
# triggered by an orchestrator function.
# Before running this sample, please:
# - create a Durable orchestration function
# - create a Durable HTTP starter function
# - add azure-functions-durable to requirements.txt
# - run pip install -r requirements.txt

import logging


def main(name: str) -> str:
    return f"Hello {name}!"



from azure.durable_functions.models.ReplaySchema import ReplaySchema
from .orchestrator_test_utils \
    import assert_orchestration_state_equals, get_orchestration_state_result, assert_valid_schema
from tests.test_utils.ContextBuilder import ContextBuilder
from azure.durable_functions.models.OrchestratorState import OrchestratorState
from azure.durable_functions.models.actions.CallSubOrchestratorAction \
    import CallSubOrchestratorAction
import azure.durable_functions as df
import azure.functions as func

app = df.DFApp(http_auth_level=func.AuthLevel.ANONYMOUS)

def generator_function(context):
    outputs = []
    task1 = yield context.call_sub_orchestrator("HelloSubOrchestrator", "Tokyo")
    task2 = yield context.call_sub_orchestrator("HelloSubOrchestrator", "Seattle")
    task3 = yield context.call_sub_orchestrator("HelloSubOrchestrator", "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs

def generator_function_call_by_function_name(context):
    outputs = []
    task1 = yield context.call_sub_orchestrator(HelloSubOrchestrator, "Tokyo")
    task2 = yield context.call_sub_orchestrator(HelloSubOrchestrator, "Seattle")
    task3 = yield context.call_sub_orchestrator(HelloSubOrchestrator, "London")

    outputs.append(task1)
    outputs.append(task2)
    outputs.append(task3)

    return outputs

@app.orchestration_trigger(context_name="context")
def HelloSubOrchestrator(context):
    return "Hello" + context

def base_expected_state(output=None, replay_schema: ReplaySchema = ReplaySchema.V1) -> OrchestratorState:
    return OrchestratorState(is_done=False, actions=[], output=output, replay_schema=replay_schema.value)


def add_hello_suborch_action(state: OrchestratorState, input_: str):
    action = CallSubOrchestratorAction(function_name='HelloSubOrchestrator', _input=input_)
    state.actions.append([action])


def add_hello_suborch_completed_events(
        context_builder: ContextBuilder, id_: int, result: str):
    context_builder.add_sub_orchestrator_started_event(name="HelloSubOrchestrator", id_=id_,input_="")
    context_builder.add_orchestrator_completed_event()
    context_builder.add_orchestrator_started_event()
    context_builder.add_sub_orchestrator_completed_event(result=result, id_=id_)

def test_tokyo_and_seattle_and_london_state():
    context_builder = ContextBuilder('test_simple_function')
    add_hello_suborch_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_suborch_completed_events(context_builder, 1, "\"Hello Seattle!\"")
    add_hello_suborch_completed_events(context_builder, 2, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_suborch_action(expected_state, 'Tokyo')
    add_hello_suborch_action(expected_state, 'Seattle')
    add_hello_suborch_action(expected_state, 'London')
    expected_state._is_done = True
    expected = expected_state.to_json()

    #assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)

def test_call_suborchestrator_by_name():
    context_builder = ContextBuilder('test_call_suborchestrator_by_name')
    add_hello_suborch_completed_events(context_builder, 0, "\"Hello Tokyo!\"")
    add_hello_suborch_completed_events(context_builder, 1, "\"Hello Seattle!\"")
    add_hello_suborch_completed_events(context_builder, 2, "\"Hello London!\"")

    result = get_orchestration_state_result(
        context_builder, generator_function_call_by_function_name)

    expected_state = base_expected_state(
        ['Hello Tokyo!', 'Hello Seattle!', 'Hello London!'])
    add_hello_suborch_action(expected_state, 'Tokyo')
    add_hello_suborch_action(expected_state, 'Seattle')
    add_hello_suborch_action(expected_state, 'London')
    expected_state._is_done = True
    expected = expected_state.to_json()

    #assert_valid_schema(result)
    assert_orchestration_state_equals(expected, result)



from typing import List

import os
from os.path import dirname

import json
import pathlib
import logging

from azure.storage.blob import BlobServiceClient
from azure.core.exceptions import ResourceExistsError

import azure.functions as func
import azure.durable_functions as df

myApp = df.DFApp(http_auth_level=func.AuthLevel.ANONYMOUS)

@myApp.route(route="orchestrators/{functionName}")
@myApp.durable_client_input(client_name="client")
async def HttpStart(req: func.HttpRequest, client):
    payload: str = json.loads(req.get_body().decode()) # Load JSON post request data
    instance_id = await client.start_new(req.route_params["functionName"], client_input=payload)

    logging.info(f"Started orchestration with ID = '{instance_id}'.")

    return client.create_check_status_response(req, instance_id)

@myApp.orchestration_trigger(context_name="context")
def E2_BackupSiteContent(context: df.DurableOrchestrationContext):
    root_directory: str = context.get_input()

    if not root_directory:
        raise Exception("A directory path is required as input")

    files = yield context.call_activity("E2_GetFileList", root_directory)
    tasks = []
    for file in files:
        tasks.append(context.call_activity("E2_CopyFileToBlob", file))
    
    results = yield context.task_all(tasks)
    total_bytes = sum(results)
    return total_bytes

connect_str = os.getenv('AzureWebJobsStorage')

@myApp.activity_trigger(input_name="rootDirectory")
def E2_GetFileList(rootDirectory):
    all_file_paths = []
    # We walk the file system
    for path, _, files in os.walk(rootDirectory):
        # We copy the code for activities and orchestrators
        if "E2_" in path:
            # For each file, we add their full-path to the list
            for name in files:
                if name == "__init__.py" or name == "function.json":
                    file_path = os.path.join(path, name)
                    all_file_paths.append(file_path)
    
    return all_file_paths

@myApp.activity_trigger(input_name="filePath")
def E2_CopyFileToBlob(filePath):
    # Create the BlobServiceClient object which will be used to create a container client
    blob_service_client = BlobServiceClient.from_connection_string(connect_str)
    
    # Create a unique name for the container
    container_name = "backups"
    
    # Create the container if it does not exist
    try:
        blob_service_client.create_container(container_name)
    except ResourceExistsError:
        pass

    # Create a blob client using the local file name as the name for the blob
    parent_dir, fname = pathlib.Path(filePath).parts[-2:] # Get last two path components
    blob_name = parent_dir + "_" + fname
    blob_client = blob_service_client.get_blob_client(container=container_name, blob=blob_name)

    # Count bytes in file
    byte_count = os.path.getsize(filePath)

    # Upload the created file
    with open(filePath, "rb") as data:
        blob_client.upload_blob(data)

    return byte_count


class EntityId:
    """EntityId.

    It identifies an entity by its name and its key.
    """

    def __init__(self, name: str, key: str):
        """Instantiate an EntityId object.

        Identifies an entity by its name and its key.

        Parameters
        ----------
        name: str
            The entity name
        key: str
            The entity key

        Raises
        ------
            ValueError: If the entity name or key are the empty string
        """
        if name == "":
            raise ValueError("Entity name cannot be empty")
        if key == "":
            raise ValueError("Entity key cannot be empty")
        self.name: str = name
        self.key: str = key

    @staticmethod
    def get_scheduler_id(entity_id: 'EntityId') -> str:
        """Produce a SchedulerId from an EntityId.

        Parameters
        ----------
        entity_id: EntityId
            An EntityId object

        Returns
        -------
        str:
            A SchedulerId representation of the input EntityId
        """
        return f"@{entity_id.name.lower()}@{entity_id.key}"

    @staticmethod
    def get_entity_id(scheduler_id: str) -> 'EntityId':
        """Return an EntityId from a SchedulerId string.

        Parameters
        ----------
        scheduler_id: str
            The SchedulerId in which to base the returned EntityId

        Raises
        ------
        ValueError:
            When the SchedulerId string does not have the expected format

        Returns
        -------
        EntityId:
            An EntityId object based on the SchedulerId string
        """
        sched_id_truncated = scheduler_id[1:]  # we drop the starting `@`
        components = sched_id_truncated.split("@")
        if len(components) != 2:
            raise ValueError("Unexpected format in SchedulerId")
        [name, key] = components
        return EntityId(name, key)

    @staticmethod
    def get_entity_id_url_path(entity_id: 'EntityId') -> str:
        """Print the the entity url path.

        Returns
        -------
        str:
            A url path of the EntityId
        """
        return f'entities/{entity_id.name}/{entity_id.key}'

    def __str__(self) -> str:
        """Print the string representation of this EntityId.

        Returns
        -------
        str:
            A SchedulerId-based string representation of the EntityId
        """
        return EntityId.get_scheduler_id(entity_id=self)



import logging,json
import os
import time
from typing import Dict
import azure.functions as func
import requests
from azureml.core import Experiment, Workspace
from azureml.core.authentication import ServicePrincipalAuthentication
from azureml.pipeline.core import PipelineRun

from ..shared.aml_helper import get_run_url_from_env, get_run_logs
from ..shared.auth_helper import get_service_principal_auth

_SUBSCRIPTION_ID_ENV_NAME = "SubscriptionId"
_RESOURCE_GROUP_NAME_ENV_NAME = "ResourceGroupName"
_AML_WORKSPACE_NAME_ENV_NAME = "AMLWorkspaceName"


def get_aml_pipeline_run_status(run_id, experiment_name, retries=3):
    
    svc_pr = get_service_principal_auth()
    workspace = Workspace(
        subscription_id=os.environ[_SUBSCRIPTION_ID_ENV_NAME],
        resource_group=os.environ[_RESOURCE_GROUP_NAME_ENV_NAME],
        workspace_name=os.environ[_AML_WORKSPACE_NAME_ENV_NAME],
        auth=svc_pr)

    experiment = Experiment(workspace, experiment_name)
    pipeline_run = PipelineRun(experiment, run_id)

    response = pipeline_run.get_status()
    return response


def main(name):
    input_args = json.loads(name)
    run_id = input_args["run_id"]
    experiment_name = input_args["experiment_name"]
    status = get_aml_pipeline_run_status(run_id,experiment_name)
    run_url = get_run_url_from_env(run_id,experiment_name)
    run_logs = get_run_logs(run_id,experiment_name)
    status_code_map = {"Finished":200,"Failed":500,"Cancelled":500}
    
    response_obj = {
        "status" : status,
        "url" : run_url,
        "logs" : run_logs,
        "status_code": status_code_map[status] if status in status_code_map else 202
    }
    return json.dumps(response_obj)
